/*
Written by Marcus Sorensen
Copyright (C) 2010-2011
learnitwithme.com

SeekMark - a threaded random I/O tester, designed to test the number of seeks 
and hence get a rough idea of the access time of a disk and the number of iops 
it can perform .  It was loosely inspired by the 'seeker' program 
(http://www.linuxinsight.com/how_fast_is_your_disk.html), when it was noticed 
that the results from seeker were very much the same for a RAID array as for a 
single disk, and a look at the code showed that it was doing one random read at 
a time, basically only triggering access to one drive and waiting for that to 
complete before continuing. What we want to see is not only the performance of 
a single spindle, but how much benefit we get on random reads as we add 
spindles.

This code has been successfully compiled and run on Linux. The source for this 
program (SeekMark) is distributed under the Clarified Artistic License. This is 
unsupported software, but please make an effort to report any bugs to me at 
<mlsorensen@mlsorensen.com>

compile example: gcc -o seekmark seekmark.c -lpthread

Versions:
0.1 - Release Candidate 1

0.3 - Release Candidate 2 - changed various report items
*/

#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/timeb.h>
#include <pthread.h>

char version[] = "SeekMark 0.1 (02/10/2011) by Marcus Sorensen";

char *file[255] = {""};//output file name
int fd;
char buf[512];
int seeks = 3000;
int numthreads = 1;
off_t size;

/* prototypes */
void seekthenread(int fd, off_t offset);
void * threadseek(void * ptr);
void usage();
void getargs(int argc, char *argv[]);

int main(int argc, char *argv[]) {
	
	//don't buffer what we print
	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);
	
	//get arguments, set things up
	getargs(argc,argv);
	
	pthread_t threads[numthreads];//create an array of threads
	errno=0;
	int i;
	struct timeb starttm,endtm; //keep track of timing on things
	long long int startsec,endsec,position,printsize;
	double startms,endms,totaltime,totalseekspersec;
	char unit[3] = "B";
	
	//test file
	fd=open(*file, O_RDONLY);
	if(fd == -1) {
		usage();
		fprintf(stderr,"\nfile %s is not readable: %s\n\n", *file,strerror(errno));
		exit(1);
	}
	
	size = lseek(fd,0,SEEK_END);
	
	if(size == -1) {
		fprintf(stderr,"\nfile %s is not readable: %s\n\n", *file,strerror(errno));
		exit(1);
	}
	
	//format display of file size
	printsize = size;
	if(size > 1048576) {
		printsize = printsize>>20;
		strcpy(unit,"MB");
	}
	else if (size > 1024) {
		printsize = printsize>>10;
		strcpy(unit,"KB");
	}
	
	printf("\nBenchmarking against %s %lld %s\n\n",*file,printsize,unit);
	
	//seed random
	srand((long int)time(NULL));
	
	//get start time
	ftime(&starttm);
	startsec = starttm.time;
	startms = starttm.millitm;
	
	//go to work	
	for(i = 0; i < numthreads; i++) {
		printf("Spawning worker %d to do %d seeks\n",i,seeks);
		pthread_create(&threads[i], NULL, threadseek, (void*)(long)i);
	}
	
	//wait for the threads to complete
	for(i = 0; i < numthreads; i++) {
		pthread_join( threads[i], NULL);
	}
	
	//get end time
	ftime(&endtm);
	endsec = endtm.time;
	endms = endtm.millitm;
	totaltime = (double)(endsec-startsec)+((endms-startms)/1000);
	totalseekspersec = ((double)seeks*numthreads)/totaltime;
	
	
	printf("\ntotal time: %.2lf, time per request(ms): %.3f\n%.2f total seeks per sec, %.2f seeks per sec per thread\n\n",totaltime,(1/totalseekspersec)*1000,totalseekspersec,totalseekspersec/numthreads);
	
}

void * threadseek(void * ptr) {
	int i;
	int threadnum = (long)ptr;
	
	//keep track of timing on things
	struct timeb starttm,endtm; 
	long long int startsec,endsec;
	double startms,endms,totaltime,seekspersec;
	

	
	//get start time
	ftime(&starttm);
	startsec = starttm.time;
	startms = starttm.millitm;
	
	//do work
	off_t position;
	for(i=0;i<(seeks);i++) {
		position = size * ((float)rand()/RAND_MAX);
		seekthenread(fd,position);
	}
	
	//get end time
	ftime(&endtm);
	endsec = endtm.time;
	endms = endtm.millitm;
	totaltime = (double)(endsec-startsec)+((endms-startms)/1000);
	seekspersec = ((double)seeks)/totaltime;
	
	
	printf("thread %d completed, time: %.2lf, %.2f seeks/sec, %.1fms per request\n",threadnum,totaltime,seekspersec,(1/seekspersec)*1000);
}

void seekthenread(int fd, off_t offset) {
	off_t seekpos = lseek(fd,offset,SEEK_SET);
	
	if( seekpos == -1) {
		printf("failed to seek to %lld/%lld: %s\n",(long long int)seekpos,(long long int)offset,strerror(errno));
	}
	
	int bytes = read(fd,buf,sizeof(buf));
	
	if(bytes =! sizeof(buf)){
		fprintf(stderr,"Error: unable to read at current position of %lld: %s\n",(long long int)seekpos,strerror(errno));
	}
}

void getargs(int argc, char *argv[]) {
	int c;
	int gotrequired = 0;
	
	while ((c = getopt (argc, argv, "f:t:s:h")) != -1){
		switch(c) {
			case 'f':
				if(strlen(optarg) > (sizeof(file)/sizeof(*file))){
					fprintf(stderr,"\nfile name too long, size should be <= %lu chars\n\n",sizeof(file)/sizeof(*file));
					exit(1);
				}
				*file = optarg;
				gotrequired = 1;
				break;
			case 't':
				if(atoi(optarg) < 1) {
					fprintf(stderr,"\nthreads should be a positive integer\n\n");
					exit(1);
				}
				numthreads = atoi(optarg);
				break;
			case 's':
				if(atoi(optarg) < 1) {
					fprintf(stderr,"\nseeks should be a positive integer\n\n");
					exit(1);
				}
				seeks = atoi(optarg);
				break;
			case 'h':
				usage();
				exit(1);
		}
	}
	
	if(gotrequired == 0) {
		usage();	
		fprintf(stderr,"\nPlease provide a filename with the -f flag\n\n");
		exit(1);
	}
}

void usage(){
	fprintf(stderr,"\n%s\n\n",version);
	fprintf(stderr,"Usage: seekmark -f FILENAME [OPTIONS]...\n");
	
	fprintf(stderr,"Defaults used are:\n");
	fprintf(stderr,"  threads:      1\n");
	fprintf(stderr,"  seeks/thread: 3000\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"OPTIONS:\n");
	fprintf(stderr,"  -f FILENAME   File to random read, device or filesystem file\n");
	fprintf(stderr,"  -t INTEGER    Number of worker threads to spawn\n");
	fprintf(stderr,"  -s INTEGER    Number of seeks to execute per thread\n");
        fprintf(stderr,"  -h            Print this help dialog and exit\n");
}

/*
                    The Clarified Artistic License

				Preamble

The intent of this document is to state the conditions under which a
Package may be copied, such that the Copyright Holder maintains some
semblance of artistic control over the development of the package,
while giving the users of the package the right to use and distribute
the Package in a more-or-less customary fashion, plus the right to make
reasonable modifications.

Definitions:

	"Package" refers to the collection of files distributed by the
	Copyright Holder, and derivatives of that collection of files
	created through textual modification.

	"Standard Version" refers to such a Package if it has not been
	modified, or has been modified in accordance with the wishes
	of the Copyright Holder as specified below.

	"Copyright Holder" is whoever is named in the copyright or
	copyrights for the package.

	"You" is you, if you're thinking about copying or distributing
	this Package.

	"Distribution fee" is a fee you charge for providing a copy
        of this Package to another party.

	"Freely Available" means that no fee is charged for the right to
        use the item, though there may be fees involved in handling the
        item.  It also means that recipients of the item may redistribute
        it under the same conditions they received it.

1. You may make and give away verbatim copies of the source form of the
Standard Version of this Package without restriction, provided that you
duplicate all of the original copyright notices and associated disclaimers.

2. You may apply bug fixes, portability fixes and other modifications
derived from the Public Domain, or those made Freely Available, or from
the Copyright Holder.  A Package modified in such a way shall still be
considered the Standard Version.

3. You may otherwise modify your copy of this Package in any way, provided
that you insert a prominent notice in each changed file stating how and
when you changed that file, and provided that you do at least ONE of the
following:

    a) place your modifications in the Public Domain or otherwise make them
    Freely Available, such as by posting said modifications to Usenet or an
    equivalent medium, or placing the modifications on a major network
    archive site allowing unrestricted access to them, or by allowing the
    Copyright Holder to include your modifications in the Standard Version
    of the Package.

    b) use the modified Package only within your corporation or organization.

    c) rename any non-standard executables so the names do not conflict
    with standard executables, which must also be provided, and provide
    a separate manual page for each non-standard executable that clearly
    documents how it differs from the Standard Version.

    d) make other distribution arrangements with the Copyright Holder.

    e) permit and encourge anyone who receives a copy of the modified Package
       permission to make your modifications Freely Available
       in some specific way.


4. You may distribute the programs of this Package in object code or
executable form, provided that you do at least ONE of the following:

    a) distribute a Standard Version of the executables and library files,
    together with instructions (in the manual page or equivalent) on where
    to get the Standard Version.

    b) accompany the distribution with the machine-readable source of
    the Package with your modifications.

    c) give non-standard executables non-standard names, and clearly
    document the differences in manual pages (or equivalent), together
    with instructions on where to get the Standard Version.

    d) make other distribution arrangements with the Copyright Holder.

    e) offer the machine-readable source of the Package, with your
       modifications, by mail order.

5. You may charge a distribution fee for any distribution of this Package.
If you offer support for this Package, you may charge any fee you choose
for that support.  You may not charge a license fee for the right to use
this Package itself.  You may distribute this Package in aggregate with
other (possibly commercial and possibly nonfree) programs as part of a
larger (possibly commercial and possibly nonfree) software distribution,
and charge license fees for other parts of that software distribution,
provided that you do not advertise this Package as a product of your own.
If the Package includes an interpreter, You may embed this Package's
interpreter within an executable of yours (by linking); this shall be
construed as a mere form of aggregation, provided that the complete
Standard Version of the interpreter is so embedded.

6. The scripts and library files supplied as input to or produced as
output from the programs of this Package do not automatically fall
under the copyright of this Package, but belong to whoever generated
them, and may be sold commercially, and may be aggregated with this
Package.  If such scripts or library files are aggregated with this
Package via the so-called "undump" or "unexec" methods of producing a
binary executable image, then distribution of such an image shall
neither be construed as a distribution of this Package nor shall it
fall under the restrictions of Paragraphs 3 and 4, provided that you do
not represent such an executable image as a Standard Version of this
Package.

7. C subroutines (or comparably compiled subroutines in other
languages) supplied by you and linked into this Package in order to
emulate subroutines and variables of the language defined by this
Package shall not be considered part of this Package, but are the
equivalent of input as in Paragraph 6, provided these subroutines do
not change the language in any way that would cause it to fail the
regression tests for the language.

8. Aggregation of the Standard Version of the Package with a commercial
distribution is always permitted provided that the use of this Package
is embedded; that is, when no overt attempt is made to make this Package's
interfaces visible to the end user of the commercial distribution.
Such use shall not be construed as a distribution of this Package.

9. The name of the Copyright Holder may not be used to endorse or promote
products derived from this software without specific prior written permission.

10. THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

				The End
*/
